"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const _ = require("underscore");
const http = require("http");
const path = require("path");
const request = require("request-promise");
const url = require("url");
const Koa = require("koa");
const KoaRouter = require("koa-router");
const KoaBodyParser = require("koa-bodyparser");
const logger = require("@nodeswork/logger");
const utils_1 = require("@nodeswork/utils");
const accounts_1 = require("./accounts");
const components_1 = require("./components");
const constants = require("./constants");
const LOG = logger.getLogger('logger');
const REQUEST_LOG = LOG;
let DEFAULT_CONFIG = {
    env: 'development',
    name: null,
    server: null,
    host: 'localhost',
    port: 28888,
    appletId: null,
    appletToken: null,
};
class Nodeswork {
    constructor(options = null) {
        this.Nodeswork = Nodeswork;
        this.Logger = components_1.Logger;
        this.Request = components_1.Request;
        this.NodesworkAccount = accounts_1.NodesworkAccount;
        this.NodesworkComponent = components_1.NodesworkComponent;
        this._options = {};
        this.initialized = false;
        this.app = new Koa();
        this.middlewares = newMiddlwares();
        this.commonMiddlewares = [];
        this.accountManager = new accounts_1.NodesworkAccountManager(this);
        this.componentManager = new components_1.NodesworkComponentManager(this);
        this.router = new KoaRouter();
        this.router
            .use(utils_1.handler.logRequestHandler(REQUEST_LOG))
            .use(utils_1.handler.internalErrorHandler(REQUEST_LOG, {
            debug: process.env.NODE_ENV != 'production'
        }))
            .use(rootHandler(this));
        if (options != null) {
            this.config(options);
        }
        this.withComponent(components_1.Logger, undefined);
    }
    /**
     * Either to config current nodeswork instance, or retrieve a config value.
     *
     * @param {NodesworkOption} options - config current nodeswork instance.
     * @param {string} options - retrieve the config value for this key.
     */
    config(options = DEFAULT_CONFIG) {
        if (_.isString(options)) {
            return this._options[options];
        }
        let env = process.env.NODE_ENV || options.env || 'development';
        let envOptions = _.chain(process.env)
            .keys()
            .filter((x) => x.startsWith('NW_'))
            .map((x) => [x.substring(3), process.env[x]]) // len('NW_') == 3
            .object()
            .value();
        let localOptions = _.omit(options, 'env');
        switch (env) {
            case 'production':
                _.extend(this._options, localOptions, envOptions, {
                    env: env,
                });
                break;
            case 'development':
            case 'test':
                _.extend(this._options, envOptions, localOptions, {
                    env: env,
                });
                break;
            default:
                throw new utils_1.NodesworkError('Unkown env when initializing nodeswork', {
                    env: env,
                    path: 'nodeswork.options.env',
                });
        }
        if (!this._options.jar) {
            this._options.jar = request.jar();
            this.requestClient = request.defaults({
                jar: this._options.jar,
                followAllRedirects: true,
                json: true
            });
        }
        return this;
    }
    /**
     * Register a new account.
     */
    withAccount(clazz, options, overwrite = false) {
        this.accountManager.register(clazz, options, overwrite);
        return this;
    }
    /**
     * Register a new component.
     */
    withComponent(clazz, options = {}, overwrite = false) {
        this.componentManager.register(clazz, options, overwrite);
        return this;
    }
    use(...middlewares) {
        Array.prototype.push.apply(this.commonMiddlewares, middlewares);
        return this;
    }
    process(...middlewares) {
        if (!this.middlewares.POST[constants.PROCESSING_URL_PATH]) {
            this.middlewares.POST[constants.PROCESSING_URL_PATH] = [];
        }
        Array.prototype.push.apply(this.middlewares.POST[constants.PROCESSING_URL_PATH], middlewares);
        return this;
    }
    view(viewPath, ...middlewares) {
        let fullPath = path.join(constants.VIEW_URL_PATH, viewPath);
        if (!this.middlewares.GET[fullPath]) {
            this.middlewares.GET[fullPath] = [];
        }
        Array.prototype.push.apply(this.middlewares.GET[fullPath], middlewares);
        return this;
    }
    action(actionPath, ...middlewares) {
        let fullPath = path.join(constants.ACTION_URL_PATH, actionPath);
        if (!this.middlewares.POST[fullPath]) {
            this.middlewares.POST[fullPath] = [];
        }
        Array.prototype.push.apply(this.middlewares.POST[fullPath], middlewares);
        return this;
    }
    /**
     * Send request using the client.
     */
    async request(options) {
        let uri;
        if (options.uri != null) {
            uri = options.uri.toString();
        }
        else {
            let aOptions = options;
            uri = aOptions.url;
            delete aOptions.url;
        }
        options.uri = url.resolve(this.config('server'), uri);
        return await this.requestClient(options);
    }
    /**
     * Set the default options for the request client.
     */
    requestDefaults(options) {
        this.requestClient = this.requestClient.defaults(options);
        return this;
    }
    async init() {
        let self = this;
        if (this.initialized) {
            return;
        }
        if (!this._options) {
            this.config();
        }
        // Check required configurations.
        for (let field of ['server', 'appletId', 'appletToken']) {
            utils_1.validator.isRequired(this._options[field], {
                message: 'Missing configurations',
                meta: {
                    path: `nodeswork.config.${field}`,
                },
            });
        }
        // Initialize Koa routes.
        let savedMiddlewares = this.middlewares;
        this.middlewares = newMiddlwares();
        await this.componentManager.initialize();
        await this.accountManager.initialize();
        await this.componentManager.initialized();
        await this.accountManager.initialized();
        _.each(savedMiddlewares, (targets, method) => {
            _.each(targets, (middlewares, path) => {
                if (self.middlewares[method][path] == null) {
                    self.middlewares[method][path] = [];
                }
                Array.prototype.push.apply(self.middlewares[method][path], middlewares);
            });
        });
        _.each(this.middlewares, (targets, method) => {
            _.each(targets, (middlewares, path) => {
                if (path.startsWith(constants.PROCESSING_URL_PATH)) {
                    self.bind.apply(self, _.flatten([
                        path, method, processHandler(self), this.commonMiddlewares,
                        middlewares
                    ]));
                }
                else if (path.startsWith(constants.VIEW_URL_PATH)) {
                    self.bind.apply(self, _.flatten([path, method, this.commonMiddlewares, middlewares]));
                }
                else if (path.startsWith(constants.ACTION_URL_PATH)) {
                    self.bind.apply(self, _.flatten([path, method, this.commonMiddlewares, middlewares]));
                }
                else {
                    self.bind.apply(self, _.flatten([path, method, this.commonMiddlewares, middlewares]));
                }
            });
        });
        // Setup app and server.
        this.app
            .use(KoaBodyParser())
            .use(this.router.routes())
            .use(this.router.allowedMethods())
            .use(utils_1.handler.uncaughtRequestHandler(REQUEST_LOG));
        this.server = http.createServer(this.app.callback());
        this.initialized = true;
    }
    async start() {
        await this.init();
        let self = this;
        await new Promise((resolve, reject) => {
            self.server.listen(self.config('port'), self.config('host') || 'localhost', (err) => {
                if (err) {
                    reject(err);
                }
                else {
                    resolve();
                }
            });
        });
    }
    bind(path, method, ...middlewares) {
        let names = _.map(middlewares, (m) => m.name || 'unknown');
        LOG.info('Bind router', {
            path: path,
            method: method,
            middlewares: names,
        });
        let params = [path];
        Array.prototype.push.apply(params, middlewares);
        this.router[method.toLowerCase()].apply(this.router, params);
    }
}
exports.Nodeswork = Nodeswork;
function logUncaughtRequests(ctx) {
    LOG.warn('Uncaught request', {
        url: ctx.request.url,
        method: ctx.request.method,
        headers: ctx.request.headers,
    });
}
function newMiddlwares() {
    return {
        GET: {},
        POST: {},
    };
}
function processHandler(nodeswork) {
    return async function processHandler(iCtx, next) {
        let ctx = iCtx;
        ctx.userApplet = ctx.request.body.userApplet;
        ctx.execution = ctx.request.body.execution;
        ctx.executionId = ctx.execution ? ctx.execution._id : null;
        utils_1.validator.isRequired(ctx.userApplet && ctx.userApplet._id, {
            message: 'userApplet is required.',
            meta: {
                responseCode: 400,
            },
        });
        ctx.accounts = await nodeswork.accountManager.parseAccounts(ctx, ctx.userApplet.accounts);
        await next();
    };
}
function rootHandler(nodeswork) {
    return async function rootHandler(ctx, next) {
        let nCtx = ctx;
        let startTime = Date.now();
        nCtx.nodeswork = nodeswork;
        nCtx.response.headers['nodeswork-processor'] = nodeswork.config('name');
        try {
            nCtx.components = nodeswork.componentManager.getComponentMap(nCtx);
            await next();
        }
        finally {
            nCtx.response.headers['nodeswork-processing-duration'] = (Date.now() - startTime);
        }
    };
}

//# sourceMappingURL=nodeswork.js.map
